const util = require('util');
import { NtwNode, NtwNodeInterface } from './ntw-node';

import { Observable, BehaviorSubject } from 'rxjs/Rx';
import 'rxjs/observable/of';
import 'rxjs/add/observable/from';
import 'rxjs/add/observable/zip';
import 'rxjs/add/operator/do';
import 'rxjs/add/operator/map';
import 'rxjs/add/operator/filter';
import 'rxjs/add/operator/mergeMap';
import 'rxjs/add/operator/switchMap';
import 'rxjs/add/operator/concatMap';
import 'rxjs/add/operator/combineLatest';
import 'rxjs/add/operator/withLatestFrom';
import 'rxjs/add/operator/bufferCount';

export class NtwProcessor {

  nodes: Map<any, NtwNodeInterface>;
  context: any;

  constructor(nodes: any, context?: any) {
    this.nodes = this.getNodes(nodes);
    this.context = context || {};
  }

  init(value: any, nodeId?: any) {
    this.context.deps = this.getDepsMap(this.nodes);
    return this.next(value, nodeId);
  }

  next(value, nodeId) {
    const node = this.getNodeById(nodeId || 0);

    if (!node) {
      return Observable.of(node);
    }

    const nextValue = this.getNextValue(node, value);

    if (!node.hasNext()) {
      return nextValue;
    }

    return Observable.create(observer => {
      const nextIds = nextValue
        .map(value => this.getNextIds(node.next, value))
        .mergeMap(arr => Observable.from(arr));

      nextIds
        .withLatestFrom(nextValue, (nextId, value) => [value, this.resolveNext(nextId, value)])
        .concatMap(values => this.next(values[0], values[1]))
        .subscribe({
          next: nextObs => observer.next(nextObs),
          error: err => console.log(err),
          complete: () => observer.complete()
        });
    });
  }

  getNodeById(nodeId) {
    return this.nodes.get(nodeId);
  }

  getNextValue(node, currentValue) {
    const nextValue = node.getNextValue(currentValue, this.context);
    this.propagateValue(node.id, nextValue);
    return this.wrapIntoObservable(nextValue);
  }

  propagateValue(id, value) {
    const deps = this.context.deps;
    if (deps && deps.has(id)) {
      deps.get(id).next(value);
    }
  }

  wrapIntoObservable(element) {
    return element instanceof Observable ? element : Observable.of(element);
  }

  getNextIds(arr, value) {
    return arr.reduce((acc, item) => {
      return acc.concat(this.resolveNext(item, value));
    }, []);
  }

  resolveNext(next, value) {
    return typeof next === 'function' ? next(value) : next;
  }

  getDepsMap(nodes) {
    const deps = new Map();
    nodes.forEach((node, key) => {
      const behaviorSubject = new BehaviorSubject(null);
      deps.set(key, behaviorSubject.filter(e => e));
    });
    return deps;
  }

  getNodes(nodes) {
    return this.everyIsFunction(nodes)
      ? this.getNodesAsLinearPipe(nodes)
      : this.getNodesAsConfigured(nodes);
  }

  everyIsFunction(arr) {
    return arr.every(item => typeof item === 'function');
  }

  getNodesAsLinearPipe(nodes) {
    return nodes.reduce((map, input, index) => {
      input = {
        id: index,
        fn: input,
        next: index + 1 < nodes.length
          ? index + 1
          : false
      };
      return this.setNewNode(map, input);
    }, new Map());
  }

  getNodesAsConfigured(nodes) {
    return nodes.reduce((map, input) => {
      return this.setNewNode(map, input);
    }, new Map());
  }

  isArrayOfFunctions(fns) {
    return Array.isArray(fns) && fns.every(fn => typeof fn === 'function');
  }

  setNewNode(map, input) {
    input.fn = this.getNodeFunctions(input.fn);
    const node = new NtwNode(input);
    return map.set(node.id, node);
  }

  getNodeFunctions(fns) {
    if (this.isArrayOfFunctions(fns)) {
      return (currentValue, context) => {
        const contextCopy = Object.assign({}, context);
        const linearNetwork = new NtwProcessor(fns, contextCopy);
        return linearNetwork.init(currentValue);
      };
    }
    return fns;
  }
}
